import re

from fabric.api import sudo, run, env


class BaseVcs(object):
    """ Core VCS api
    """

    TAG = 'base'
    NAME = 'Base VCS'

    def __init__(self, project_root, use_sudo=None, code_dir=None, **kwargs):
        self.project_root = project_root
        self.code_dir = code_dir
        self.use_sudo = use_sudo

        if self.use_sudo is None:
            self.use_sudo = getattr(env, 'use_sudo', False)

        if self.code_dir is None:
            self.code_dir = getattr(env, 'code_dir', None)

            if self.code_dir is None:
                raise EnvironmentError('%s: Please provide code_dir (via `init_kwargs` or `env`)' % self.NAME)

    def repo_url(self):
        """ Retrieve Url of the remote repository (origin|default). If remote url can't be determined None is returned.

        :return: remote url
        :rtype: str|None
        """
        raise NotImplementedError  # pragma: no cover

    def clone(self, revision=None):
        """ Clones the project to a target machine. Will abort if something goes wrong.

        :param revision: Can be used to specify a branch or commit that should be activated after cloning.
        """
        raise NotImplementedError  # pragma: no cover

    def get_branch(self):
        """ Get current active branch in target machine.

        :return: current active branch
        :rtype: string
        """
        raise NotImplementedError  # pragma: no cover

    def pull(self):
        """ Update the cloned repository on the target machine without changing
            the working copy. Internally this is done via `git fetch` or `hg pull`.
        """
        raise NotImplementedError  # pragma: no cover

    def update(self, revision=''):
        """ Update the target to specified revision or tip of currently active branch if revision is omitted.

        :param revision: Specific revision to update to
        """
        raise NotImplementedError  # pragma: no cover

    def deployment_list(self, revision=''):
        """ List revisions to apply/un-apply when updating to given revision (if not specified defaults to tip of currently active branch).

            The returned dict can have the following keys:

            - forwards:     If there are revisions to apply, this value will contain a list with information about each commit
            - backwards:    If there are revisions to un-apply, this value will contain a list with information about each commit
            - revset:       If there are some revisions to apply/un-apply this will contain a string that can be passed on to changed_files
            - message:      If no revisions are to be applied or something else is wrong, this will contain this information

        :param revision: Specific revision to diff against
        :rtype: dict
        """
        raise NotImplementedError  # pragma: no cover

    def get_revset_log(self, revs):
        """ Returns lines returned by hg|git log as a list.

        :param revs: Revision set passed onto `hg|git log`
        :return: list of commits (lines are in the following format: `commit_hash branch author description`)
        :rtype: list
        """

        raise NotImplementedError  # pragma: no cover

    def version(self):
        """ Get the commit id, branch, message and author active on the target machine

        :return: (commit_id, branch, message, author)
        :rtype: tuple
        """

        raise NotImplementedError  # pragma: no cover

    def remote_cmd(self, *args, **kwargs):
        if not self.use_sudo:  # pragma: no cover
            return run(*args, **kwargs)

        else:
            return sudo(*args, **kwargs)  # pragma: no cover

    def _changed_files(self, revision_set):
        raise NotImplementedError  # pragma: no cover

    def changed_files(self, revision_set, filter_re=None):
        """ Returns list of files that changed in the given revset, optionally filtered by the given regex or list of regex values.

            Each returned item is a combination of the action and the file name separated by space:
            > ['A test.tx', 'M foo.bar', 'R hello_world']

        :param revision_set: Revision set to use when building changed file list
        :param filter_re: optional regex filter (or list of regex values)
        :return: files changed
        :rtype: list
        """
        result = self._changed_files(revision_set)

        if filter_re:
            def finder(pattern):
                regex = re.compile(pattern)

                return filter(lambda filename: regex.search(filename), result)

            if isinstance(filter_re, (list, tuple)):
                full_result = []

                for reg in filter_re:
                    for res in finder(reg):
                        full_result.append(res)

                result = full_result

            else:
                result = finder(filter_re)

        return result

    @classmethod
    def get_revset(cls, x, y):
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def get_revisions(cls, x):
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def detect(cls, project_root, **init_kwargs):
        """ Detect if the current VCS implementation can be used for the
            active project.

            :param project_root: Root directory of the current project
            :rtype: bool
        """

        raise NotImplementedError  # pragma: no cover



import os

from fabric.api import cd, hide, lcd, local, abort

from .base import BaseVcs


class Mercurial(BaseVcs):
    TAG = 'hg'
    NAME = 'Mercurial'

    def version(self):
        with cd(self.code_dir), hide('running'):
            commit_id, branch = self.remote_cmd('hg id -nb', quiet=True).split()

            separator = ':|:|:'
            c_hash, author, message = self.remote_cmd(("hg log --template "
                                                       "'{node|short}%(sep)s{author}%(sep)s{desc|firstline}\\n' -r %(id)s") % dict(
                sep=separator,
                id=commit_id,
            )).split(separator)

            commit_id = '%s:%s' % (commit_id, c_hash)

        return commit_id, branch, message, author

    def repo_url(self):
        with lcd(self.project_root):
            try:
                result = local('hg paths default', capture=True)

            except SystemExit as e:
                if e.code == 1:
                    result = None

                else:
                    raise  # pragma: no cover

            return result or None

    def clone(self, revision=None):
        repo_url = self.repo_url()

        if not repo_url:
            abort('Repo url was not found')  # pragma: no cover

        self.remote_cmd('hg clone %s %s' % (repo_url, self.code_dir))

        if revision:
            self.update(revision)

    def get_branch(self):
        with cd(self.code_dir), hide('running'):
            return self.remote_cmd('hg id -b', quiet=True).strip()

    def pull(self):
        with cd(self.code_dir):
            self.remote_cmd('hg pull')

    def update(self, revision=''):
        with cd(self.code_dir):
            self.remote_cmd('hg update %s' % revision)

    def get_revset_log(self, revs):
        with cd(self.code_dir):
            result = self.remote_cmd("hg log --template '{rev}:{node|short} {branch} {author} {desc|firstline}\\n' -r '%s'" % revs)

            if not result:
                return []

            return list(filter(lambda y: y, [x.strip() for x in (result.split('\n') or [])]))

    def deployment_list(self, revision=''):
        with cd(self.code_dir), hide('running', 'stdout'):
            # First lets pull
            self.pull()

            revision_set = self.get_revset('.', revision)
            revisions = self.get_revset_log(revision_set)

            if len(revisions) > 1:
                # Target is forward of the current rev
                return {'forwards': self.get_revisions(revisions), 'revset': revision_set}

            elif len(revisions) == 1:
                # Current rev is the same as target
                return {'message': "Already at target revision"}

            # Check if target is backwards of the current rev
            revision_set = self.get_revset(revision, '.')
            revisions = self.get_revset_log(revision_set)

            if revisions:
                return {'backwards': list(reversed(self.get_revisions(revisions))), 'revset': revision_set}

            else:
                return {'message': "Target revision is not related to the current revision"}  # pragma: no cover

    def _changed_files(self, revision_set):
        with cd(self.code_dir):
            result = self.remote_cmd("hg status --rev '%s'" % revision_set, quiet=True).splitlines()

            return result

    @classmethod
    def get_revset(cls, x, y):
        assert x or y

        if x and y:
            # All revisions that are descendants of the current revision and ancestors of the target revision
            #  (inclusive), but not the current revision itself
            return '%s::%s' % (x, y)
        else:
            # All revisions that are in the current branch, are descendants of the current revision and are not the
            #  current revision itself.
            return 'branch(p1()) and %s::%s' % (x or '', y or '')

    @classmethod
    def get_revisions(cls, x):
        return x[1:]

    @classmethod
    def detect(cls, project_root, **init_kwargs):
        return os.path.exists(os.path.join(project_root, '.hg'))



import os

from fabric import colors
from fabric.api import cd, hide, abort, local, lcd

from .base import BaseVcs


class Git(BaseVcs):
    TAG = 'git'
    NAME = 'Git'

    def repo_url(self):
        with lcd(self.project_root):
            try:
                result = local('git config --get remote.origin.url', capture=True)

            except SystemExit as e:
                if e.code == 1:
                    result = None

                else:
                    raise  # pragma: no cover

            return result or None

    def clone(self, revision=None):
        repo_url = self.repo_url()

        if not repo_url:
            abort('Repo url was not found')  # pragma: no cover

        self.remote_cmd('git clone %(repo)s %(dir)s' % dict(
            repo=repo_url,
            dir=self.code_dir,
        ))

        # update to a specific version
        if revision:
            self.update(revision)

    def version(self):
        with cd(self.code_dir), hide('running'):
            sep = ':|:|:'

            commit_id, message, author = self.remote_cmd(("git --no-pager log -n 1 --oneline "
                                                         "--format='%%h%(sep)s%%s%(sep)s%%an <%%ae>'") % dict(sep=sep)).split(sep)

            branch = self.get_branch('HEAD')

        return commit_id, branch, message, author

    def get_branch(self, commit_id='HEAD', ambiguous=False):
        with cd(self.code_dir), hide('running'):
            try:
                branch = self.remote_cmd('git symbolic-ref --short -q %s' % commit_id)

            except SystemExit as e:
                if e.code == 1:
                    # Previous command will exit with exit code 1 if in detached
                    # head state.
                    branch = None
                else:
                    raise  # pragma: no cover

            if not branch:
                # Lets attempt to figure out the branch
                res = self.remote_cmd("git --no-pager log -n 1 --oneline --pretty=%%d %s" % commit_id).strip(' ()').split(',')
                res = [y for y in [x.strip() for x in res] if y]

                # Figure out the matching branch name
                valid_branches = []
                for candidate in res:
                    if 'HEAD' in candidate:
                        continue

                    if candidate.startswith('origin/'):
                        candidate = candidate[7:]
                        valid_branches.append(candidate)

                # in some cases even the second command won't be
                #  able to figure out the branch
                if not valid_branches:
                    valid_branches = self.git_what_branch(commit_id)

                if not valid_branches:
                    # No compatible branch found. Oh well lets just abort...
                    abort('Could not figure out remote branch (for %s)' % commit_id)  # pragma: no cover

                elif len(valid_branches) > 1:  # pragma: no cover
                    if ambiguous:
                        return '|'.join(valid_branches)

                    # Ask the user which one is valid
                    print(colors.yellow('Could not automatically determine remote git branch '
                                        '(for %s), please pick the correct value' % commit_id))
                    print('Candidates are (use 0 to abort): %s' % (
                        ', '.join(['%d: %s' % (i + 1, x) for i, x in enumerate(valid_branches)])
                    ))

                    value = None
                    while value not in range(len(valid_branches) + 1):
                        value = raw_input('Select value: ')

                        try:
                            value = int(value)

                        except (TypeError, ValueError):
                            value = None

                    if value == 0:
                        abort('Cancel by user')

                    else:
                        return valid_branches[value - 1]

                else:
                    return valid_branches[0]

        return branch

    def pull(self):
        with cd(self.code_dir):
            self.remote_cmd('git fetch origin')

    def update(self, revision=''):
        if not revision:
            revision = 'origin/%s' % self.get_branch()

        with cd(self.code_dir):
            self.remote_cmd('git checkout %s' % revision)

    def get_all_branches(self, remote=True):
        with cd(self.code_dir):
            all_branches = self.remote_cmd('git --no-pager branch%s --color=never' % ' -r' if remote else ' -a')
            all_branches = [x.strip() for x in all_branches.splitlines(False)]

            return set(list(map(self.normalize_branch, filter(lambda y: y, all_branches))))

    def get_commit_id(self):
        with cd(self.code_dir):
            return self.remote_cmd('git --no-pager log -n 1 --oneline --pretty=%h').strip()

    def git_what_branch(self, commit_id):
        if commit_id.lower() == 'head':
            commit_id = self.get_commit_id()

        with cd(self.code_dir):
            all_branches = self.get_all_branches()
            valid_branches = []

            for branch in all_branches:
                try:
                    commit_log = self.remote_cmd(('git --no-pager log --oneline '
                                                  '%(branch)s origin/%(branch)s --pretty=%%h | grep %(commit)s') % dict(
                        branch=branch,
                        commit=commit_id,
                    ))

                    if commit_log and commit_id in commit_log:  # pragma: no branch
                        valid_branches.append(branch)

                except SystemExit as e:
                    if e.code == 1:
                        continue

                    else:
                        raise  # pragma: no cover

        return valid_branches

    def deployment_list(self, revision=''):
        if not revision:
            revision = 'origin/%s' % self.get_branch()

        with cd(self.code_dir), hide('running', 'stdout'):
            # First lets pull
            self.pull()

            revision_set = self.get_revset(' ', revision)
            revisions = self.get_revset_log(revision_set)

            if len(revisions) > 0:
                # Target is forward of the current rev
                return {'forwards': self.get_revisions(revisions), 'revset': revision_set}

            # Check if target is backwards of the current rev
            revision_set = self.get_revset(revision, ' ')
            revisions = self.get_revset_log(revision_set)

            if revisions:
                return {'backwards': list(reversed(self.get_revisions(revisions))), 'revset': revision_set}

            else:
                return {'message': "Already at target revision"}

    def _changed_files(self, revision_set):
        with cd(self.code_dir):
            result = self.remote_cmd("git --no-pager diff --name-status %s" % revision_set, quiet=True).splitlines()

            return map(lambda x: x.replace('\t', ' '), result)

    def get_revset_log(self, revs):
        with cd(self.code_dir):
            result = self.remote_cmd("git --no-pager log %s --oneline --format='%%h %%(branch)s %%an <%%ae> %%s'" % revs).strip()

            if result:
                result = result.split('\n')

                return map(self.log_add_branch, filter(lambda y: y, [x.strip() for x in result]))

            return []

    def log_add_branch(self, line, needle='branch'):
        if not line:
            return line  # pragma: no cover

        commit_hash = line.split()[0]

        if not commit_hash:
            return line  # pragma: no cover

        return line % {needle: self.get_branch(commit_hash, ambiguous=True)}

    @classmethod
    def get_revset(cls, x, y):
        assert x and y, NotImplementedError('With git, get_revset should always have x and y')

        return '%s..%s' % (x, y)

    @classmethod
    def get_revisions(cls, x):
        return list(reversed(x))

    @classmethod
    def detect(cls, project_root, **init_kwargs):
        return os.path.exists(os.path.join(project_root, '.git'))

    @staticmethod
    def normalize_branch(branch):
        return branch.replace('origin/', '').replace('HEAD', '').replace('->', '').strip('/').strip()



hammer_name = 'tg-hammer'
hammer_description = 'Helpers for fabric based deployments.'
hammer_version = '0.0.3'

__name__ = hammer_name
__description__ = hammer_description
__version__ = hammer_version



#!/usr/bin/env python
import uuid

import hammer

from pip.req import parse_requirements

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


requirements = [str(ir.req) for ir in parse_requirements('requirements/base.txt', session=uuid.uuid1())]
test_requirements = [str(ir.req) for ir in parse_requirements('requirements/development.txt', session=uuid.uuid1())]

setup(
    name=hammer.__name__,
    version=hammer.__version__,
    description=hammer.__description__,
    long_description=open('README.md').read(),
    author='Thorgate',
    author_email='hi@thorgate.eu',
    url='https://github.com/thorgate/tg-hammer',
    packages=[
        'hammer',
    ],
    include_package_data=True,
    install_requires=requirements,
    test_suite='py.test',
    tests_require=test_requirements,
    license="BSD",
    keywords='tg-hammer hammer fabric vcs git hg',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
    ],
)



from .git import Git
from .hg import Mercurial


class VcsProxy(object):
    VCS_HANDLERS = [
        Mercurial,
        Git
    ]

    def __init__(self, project_root, **init_kwargs):
        self._data = {
            'project_root': project_root,
            'init_kwargs': init_kwargs,
        }

        self._real = 'VcsProxy'

    @classmethod
    def init(cls, project_root, **init_kwargs):
        """ Detect the vcs type the project uses and initialize
            the correct handler to use internally.

            :param project_root: Root directory of project
            :rtype: hammer.vcs.base.BaseVcs
        """

        return VcsProxy(project_root=project_root, **init_kwargs)

    @classmethod
    def detect(cls, project_root, **init_kwargs):
        for handler_cls in cls.VCS_HANDLERS:
            res = handler_cls.detect(project_root, **init_kwargs)

            # Found a match
            if res:
                return handler_cls

        handlers = ', '.join([x.TAG for x in cls.VCS_HANDLERS])
        raise EnvironmentError('No suitable VCS type detected (tried %s)' % handlers)

    # ============
    # Proxy logic
    # ============

    def __getattribute__(self, name):
        real = object.__getattribute__(self, "_real")

        if real == 'VcsProxy':
            data = object.__getattribute__(self, "_data")

            cls = VcsProxy.detect(project_root=data['project_root'], **data['init_kwargs'])
            real = cls(project_root=data['project_root'], **data['init_kwargs'])

            self._real = real

        return getattr(real, name)

    def __str__(self):  # pragma: no cover
        return str(object.__getattribute__(self, "_real"))

    def __repr__(self):  # pragma: no cover
        return repr(object.__getattribute__(self, "_real"))
